@@ -3,22 +3,26 @@
 # See also LICENSE.txt
 
 from icemac.addressbook.i18n import MessageFactory as _
+import datetime
+import decimal
 import gocept.reference.field
 import icemac.ab.importer.browser.wizard.base
 import icemac.ab.importer.interfaces
 import icemac.addressbook.interfaces
 import persistent.mapping
+import time
 import z3c.form.field
 import zc.sourcefactory.contextual
 import zope.component
-import zope.event
 import zope.interface
 import zope.schema
 import zope.security.proxy
-import zope.site.hooks
 
 
 NONE_REPLACEMENT = object()
+TRUE_VALUES = ['yes', 'true']
+FALSE_VALUES = ['no', 'false']
+DATETIME_FORMAT = '%Y-%m-%d %H:%M'
 
 
 def get_reader(session):
@@ -87,6 +91,64 @@ def text_field(value, field):
     return value.strip()
 
 
+@zope.component.adapter(None, zope.schema.interfaces.IInt)
+@zope.interface.implementer(IFieldValue)
+def int_field(value, field):
+    "Adapter for `Int` fields."
+    if value is None:
+        # We can't return None, as this means that the adapter can't adapt.
+        return NONE_REPLACEMENT
+    try:
+        return int(float(value))
+    except (TypeError, ValueError):
+        return value
+
+
+@zope.component.adapter(None, zope.schema.interfaces.IDecimal)
+@zope.interface.implementer(IFieldValue)
+def decimal_field(value, field):
+    "Adapter for `Decimal` fields."
+    if value is None:
+        # We can't return None, as this means that the adapter can't adapt.
+        return NONE_REPLACEMENT
+    try:
+        return decimal.Decimal(value)
+    except (TypeError, ValueError, decimal.InvalidOperation):
+        return value
+
+
+@zope.component.adapter(None, zope.schema.interfaces.IBool)
+@zope.interface.implementer(IFieldValue)
+def bool_field(value, field):
+    "Adapter for `Bool` fields."
+    if value is None:
+        # We can't return None, as this means that the adapter can't adapt.
+        return NONE_REPLACEMENT
+    if not isinstance(value, unicode):
+        return value # produces an error in validation
+    value = value.lower()
+    if value in TRUE_VALUES:
+        return True
+    if value in FALSE_VALUES:
+        return False
+    return value
+
+
+@zope.component.adapter(None, zope.schema.interfaces.IDatetime)
+@zope.interface.implementer(IFieldValue)
+def datetime_field(value, field):
+    "Adapter for `Datetime` fields."
+    if value is None:
+        # We can't return None, as this means that the adapter can't adapt.
+        return NONE_REPLACEMENT
+    try:
+        time_tuple = time.strptime(value.strip(), DATETIME_FORMAT)
+    except (TypeError, ValueError):
+        return value
+    else:
+        return datetime.datetime(*time_tuple[:5])
+
+
 @zope.component.adapter(None, zope.schema.interfaces.IURI)
 @zope.interface.implementer(IFieldValue)
 def uri_field(value, field):
@@ -98,7 +160,7 @@ def uri_field(value, field):
 
 @zope.component.adapter(None, zope.schema.interfaces.IChoice)
 @zope.interface.implementer(IFieldValue)
-def country_or_state_field(value, field):
+def country_field(value, field):
     if value is None:
         # We can't return None, as this means that the adapter can't adapt.
         return NONE_REPLACEMENT
@@ -143,14 +205,18 @@ def simple_invalid(field, exc):
 @zope.interface.implementer(IErrorMessage)
 def choice_constraint_not_satisfield(field, exc):
     value = exc.args[0]
+    if zope.schema.interfaces.IVocabulary.providedBy(field.source):
+        allowed = [x.value for x in field.source]
+    else:
+        allowed = [str(x) for x in field.source.factory.getValues()]
     return _(u'Value %s not allowed. Allowed values: %s') % (
-        value, ', '.join(str(x) for x in field.source.factory.getValues()))
+        value, ', '.join(allowed))
 
 
 @zope.component.adapter(zope.schema.interfaces.IChoice,
                         zope.schema.interfaces.ConstraintNotSatisfied)
 @zope.interface.implementer(IErrorMessage)
-def country_or_state_constraint_not_satisfield(field, exc):
+def country_constraint_not_satisfield(field, exc):
     value = exc.args[0]
     titles = [_(u'%s resp. %s') % (x.token, field.source.factory.getTitle(x))
               for x in field.source.factory.getValues()]
@@ -165,6 +231,38 @@ def date_wrong_type(field, exc):
     value = exc.args[0]
     return _(u'%s is no valid date.') % value
 
+@zope.component.adapter(zope.schema.interfaces.IDatetime,
+                        zope.schema.interfaces.WrongType)
+@zope.interface.implementer(IErrorMessage)
+def datetime_wrong_type(field, exc):
+    value = exc.args[0]
+    return _(u'%s is no valid datetime. Must match to format string %r.') % (
+        value, DATETIME_FORMAT)
+
+@zope.component.adapter(zope.schema.interfaces.IInt,
+                        zope.schema.interfaces.WrongType)
+@zope.interface.implementer(IErrorMessage)
+def int_wrong_type(field, exc):
+    value = exc.args[0]
+    return _(u'%s is not a valid integer number.') % value
+
+
+@zope.component.adapter(zope.schema.interfaces.IDecimal,
+                        zope.schema.interfaces.WrongType)
+@zope.interface.implementer(IErrorMessage)
+def decimal_wrong_type(field, exc):
+    value = exc.args[0]
+    return _(u'%s is not a valid decimal number.') % value
+
+
+@zope.component.adapter(zope.schema.interfaces.IBool,
+                        zope.schema.interfaces.WrongType)
+@zope.interface.implementer(IErrorMessage)
+def bool_wrong_type(field, exc):
+    value = exc.args[0]
+    return _(u'Value %s not allowed. Allowed values: %s') % (
+        value, ', '.join(TRUE_VALUES + FALSE_VALUES))
+
 
 @zope.component.adapter(None, zope.schema.interfaces.ConstraintNotSatisfied)
 @zope.interface.implementer(IErrorMessage)
@@ -179,20 +277,19 @@ def index_error(field, exc):
     return _(u'Not enough data fields in row.')
 
 
-def render_error(interface, field_name, exc):
+def render_error(entity, field_name, exc):
     "Render the error text using the error render adapters."
-    obj_title = icemac.ab.importer.browser.wizard.base.\
-        getImportMappingRowForInterface(interface)['title']
+    obj_title = entity.title
 
     if field_name is None:
         field = ''
         title = obj_title
-        # need to set the field name to the default here, as
-        # queryMultiAdapter call with name=None behaves strange, it
-        # seems to delete all adapters matching objects and interface.
+        # Need to set the field name to the default here, as a
+        # queryMultiAdapter call with name=None behaves strange: it
+        # seems to delete all adapters matching object and interface.
         field_name = u''
     else:
-        field = interface[field_name]
+        field = entity.getField(field_name)
         title = '%s - %s' % (obj_title, field.title)
 
     # try named adapter first
@@ -205,9 +302,10 @@ def render_error(interface, field_name, exc):
 
     return u'%s: %s' % (title, message)
 
+
 class KeywordBuilder(object):
 
-    KEYWORD_FIELD_NAME = "person-0.keywords"
+    KEYWORD_FIELD_NAME = "IcemacAddressbookPersonPerson-0.keywords"
     keyword_index = None
     keywords = None
 
@@ -215,8 +313,8 @@ class KeywordBuilder(object):
         """Expects a mapping between name of the field in the address book and
         field index in import file.
 
-        Example: person-0.first_name --> 0
-                 home_page_address-2.notes --> 11
+        Example: IcemacAddressbookPersonPerson-0.first_name --> 0
+                 IcemacAddressbookAddressHomePageAddress-2.notes --> 11
         """
         if self.KEYWORD_FIELD_NAME in user_data:
             self.keyword_index = user_data[self.KEYWORD_FIELD_NAME]
@@ -247,8 +345,8 @@ class ImportObjectBuilder(object):
         field index in import file and
         .
 
-        Example: person-0.first_name --> 0
-                 home_page_address-3.notes --> 11
+        Example: IcemacAddressbookPersonPerson-0.first_name --> 0
+                 IcemacAddressbookAddressHomePageAddress-3.notes --> 11
 
         Stores the address book field names in dictionaries on
         attributes mapping to the index in the import file.
@@ -259,9 +357,12 @@ class ImportObjectBuilder(object):
         """
         self.address_book = address_book
         self.entries_number = entries_number
-        for row in icemac.ab.importer.browser.wizard.base.import_mapping:
+        self.import_entities = (
+            icemac.ab.importer.browser.wizard.base.get_import_entities())
+        for entity in self.import_entities:
             for index in xrange(self.entries_number):
-                setattr(self, "%s-%s" % (row['prefix'], index), {})
+                key = "%s-%s" % (entity.name, index)
+                setattr(self, key, {})
         for field_desc, index in user_data.iteritems():
             if index is None:
                 continue # field was not selected for import
@@ -273,20 +374,26 @@ class ImportObjectBuilder(object):
 
         data ... import data row, mapping between field index and value."""
         self.errors = set()
-        person = self._create('person-0', self.address_book, data, True)
-        self._validate(
-            icemac.ab.importer.browser.wizard.base.person_mapping['interface'],
-            person)
-        for address in icemac.addressbook.address.address_mapping:
+        person_entity = zope.component.getUtility(
+            icemac.addressbook.interfaces.IEntity,
+            name='icemac.addressbook.person.Person')
+        person = self._create('%s-0' % person_entity.name, self.address_book,
+                              data, True)
+        self._validate(person_entity, person)
+        for entity in self.import_entities:
+            if entity.class_name == 'icemac.addressbook.person.Person':
+                # already handled
+                continue
             for index in xrange(self.entries_number):
-                prefix = "%s-%s" % (address['prefix'], index)
+                prefix = "%s-%s" % (entity.name, index)
                 main_entry = (index == 0)
                 obj = self._create(prefix, person, data, main_entry)
                 if main_entry:
                     # set the created address as main address of its kind
-                    setattr(person, 'default_'+address['prefix'], obj)
+                    icemac.addressbook.person.get_default_field(
+                        entity.interface).set(person, obj)
                 if obj is not None:
-                    self._validate(address['interface'], obj)
+                    self._validate(entity, obj)
         return person, sorted(list(self.errors))
 
     def _create(self, prefix, parent, data, creation_required):
@@ -301,20 +408,19 @@ class ImportObjectBuilder(object):
             return
 
         # instantiate the object
-        row_prefix, row_index = prefix.split('-')
-        row = (
-            icemac.ab.importer.browser.wizard.base.
-            getImportMappingRowForPrefix(row_prefix))
-        obj = icemac.addressbook.utils.create_obj(row['class_'])
+        entity_name, row_index = prefix.split('-')
+        entities = zope.component.getUtility(
+            icemac.addressbook.interfaces.IEntities)
+        entity = entities.getEntity(entity_name)
+        obj = icemac.addressbook.utils.create_obj(entity.getClass())
 
         # set the values
-        iface = row['interface']
         for field_name, index in field_mapping.iteritems():
-            field = iface[field_name]
+            field = entity.getField(field_name)
             try:
                 value = data[index]
             except IndexError, e:
-                self.errors.add(render_error(iface, None, e))
+                self.errors.add(render_error(entity, None, e))
             else:
                 # try named converter first
                 conv_value = zope.component.queryMultiAdapter(
@@ -325,29 +431,35 @@ class ImportObjectBuilder(object):
                         (value, field), IFieldValue)
                 if conv_value is NONE_REPLACEMENT:
                     conv_value = None
+                context = field.interface(obj)
                 try:
-                    field.set(obj, conv_value)
+                    field.set(context, conv_value)
                 except zope.interface.Invalid, e:
-                    self.errors.add(render_error(iface, field_name, e))
+                    self.errors.add(render_error(entity, field_name, e))
 
         icemac.addressbook.utils.add(parent, obj)
         return obj
 
-    def _validate(self, interface, obj):
-        for field_name, exc in zope.schema.getValidationErrors(interface, obj):
-            self.errors.add(render_error(interface, field_name, exc))
+    def _validate(self, entity, obj):
+        for field_name, field in entity.getFieldsInOrder():
+            context = field.interface(obj)
+            field = field.bind(context)
+            value = field.get(context)
+            try:
+                field.validate(value)
+            except zope.schema.ValidationError, exc:
+                self.errors.add(render_error(entity, field_name, exc))
 
 
 class FieldsGroup(z3c.form.group.Group):
     "Fields grouped by object."
 
-    def __init__(self, context, request, parent, interface, label, prefix):
+    def __init__(self, context, request, parent, entity, label, prefix):
         super(FieldsGroup, self).__init__(context, request, parent)
         self.label = label
         self.prefix = prefix
         fields = []
-        for field_name in zope.schema.getFieldNamesInOrder(interface):
-            field = interface[field_name]
+        for field_name, field in entity.getFieldsInOrder():
             choice = zope.schema.Choice(
                 title=field.title, description=field.description,
                 required=False, source=import_fields)
@@ -375,8 +487,10 @@ class MapFields(z3c.form.group.GroupForm,
         session = self.getContent()
         request = self.request
         self.groups = []
-        for row in icemac.ab.importer.browser.wizard.base.import_mapping:
-            if row['prefix'] == 'person':
+        import_entities = (
+            icemac.ab.importer.browser.wizard.base.get_import_entities())
+        for entity in import_entities:
+            if entity.class_name  == 'icemac.addressbook.person.Person':
                 entries_number = 1
                 main_prefix = ''
             else:
@@ -387,11 +501,10 @@ class MapFields(z3c.form.group.GroupForm,
                     row_title_prefix = main_prefix
                 else:
                     row_title_prefix = _(u'other')
-                title = row_title_prefix + ' ' + row['title']
-                prefix = '%s-%s' % (row['prefix'], index)
+                title = row_title_prefix + ' ' + entity.title
+                prefix = '%s-%s' % (entity.name, index)
                 self.groups.append(
-                    FieldsGroup(session, request, self, row['interface'],
-                                title, prefix))
+                    FieldsGroup(session, request, self, entity, title, prefix))
 
     @property
     def fields(self):
